@@ -20,6 +20,8 @@ import logging
 import multiprocessing
 import os
 import stat
+import subprocess
+import sys
 import tempfile
 
 import fixtures
@@ -29,6 +31,8 @@ import six
 
 from oslo_concurrency import processutils
 from oslotest import mockpatch
+
+
 PROCESS_EXECUTION_ERROR_LOGGING_TEST = """#!/bin/bash
 exit 41"""
 
@@ -40,6 +44,12 @@ echo onstdout --password='"secret"'
 echo onstderr --password='"secret"' 1>&2
 exit 38"""
 
+# This byte sequence is undecodable from most encoding
+UNDECODABLE_BYTES = b'[a\x80\xe9\xff]'
+
+TRUE_UTILITY = (sys.platform.startswith('darwin') and
+                '/usr/bin/true' or '/bin/true')
+
 
 class UtilsTest(test_base.BaseTestCase):
     # NOTE(jkoelker) Moar tests from nova need to be ported. But they
@@ -60,6 +70,53 @@ class UtilsTest(test_base.BaseTestCase):
                                                         mock_cpu_count):
         self.assertEqual(1, processutils.get_worker_count())
 
+    def test_execute_with_callback(self):
+        on_execute_callback = mock.Mock()
+        on_completion_callback = mock.Mock()
+        processutils.execute(TRUE_UTILITY)
+        self.assertEqual(0, on_execute_callback.call_count)
+        self.assertEqual(0, on_completion_callback.call_count)
+
+        processutils.execute(TRUE_UTILITY, on_execute=on_execute_callback,
+                             on_completion=on_completion_callback)
+        self.assertEqual(1, on_execute_callback.call_count)
+        self.assertEqual(1, on_completion_callback.call_count)
+
+    @mock.patch.object(subprocess.Popen, "communicate")
+    def test_execute_with_callback_and_errors(self, mock_comm):
+        on_execute_callback = mock.Mock()
+        on_completion_callback = mock.Mock()
+
+        def fake_communicate(*args):
+            raise IOError("Broken pipe")
+
+        mock_comm.side_effect = fake_communicate
+
+        self.assertRaises(IOError,
+                          processutils.execute,
+                          TRUE_UTILITY,
+                          on_execute=on_execute_callback,
+                          on_completion=on_completion_callback)
+        self.assertEqual(1, on_execute_callback.call_count)
+        self.assertEqual(1, on_completion_callback.call_count)
+
+    def test_execute_with_preexec_fn(self):
+        # NOTE(dims): preexec_fn is set to a callable object, this object
+        # will be called in the child process just before the child is
+        # executed. So we cannot pass share variables etc, simplest is to
+        # check if a specific exception is thrown which can be caught here.
+        def preexec_fn():
+            raise processutils.InvalidArgumentError()
+
+        processutils.execute(TRUE_UTILITY)
+
+        expected_exception = (processutils.InvalidArgumentError if six.PY2
+                              else subprocess.SubprocessError)
+        self.assertRaises(expected_exception,
+                          processutils.execute,
+                          TRUE_UTILITY,
+                          preexec_fn=preexec_fn)
+
 
 class ProcessExecutionErrorTest(test_base.BaseTestCase):
 
@@ -168,6 +225,13 @@ exit 1
                           processutils.execute,
                           '/usr/bin/env', 'false', check_exit_code=True)
 
+    def test_check_cwd(self):
+        tmpdir = tempfile.mkdtemp()
+        out, err = processutils.execute('/usr/bin/env',
+                                        'sh', '-c', 'pwd',
+                                        cwd=tmpdir)
+        self.assertIn(tmpdir, out)
+
     def test_check_exit_code_list(self):
         processutils.execute('/usr/bin/env', 'sh', '-c', 'exit 101',
                              check_exit_code=(101, 102))
@@ -257,7 +321,7 @@ grep foo
 
         self.assertEqual(attempts if attempts else 1, mock.mock.call_count)
         self.assertIn('Got an OSError', fixture.output)
-        self.assertIn('errno: 11', fixture.output)
+        self.assertIn('errno: %d' % errno.EAGAIN, fixture.output)
         self.assertIn("'/usr/bin/env false'", fixture.output)
 
     def test_logging_on_communicate_error_1(self):
@@ -294,6 +358,19 @@ grep foo
         env_vars = {'SUPER_UNIQUE_VAR': 'The answer is 42'}
 
         out, err = processutils.execute('/usr/bin/env', env_variables=env_vars)
+        self.assertIsInstance(out, str)
+        self.assertIsInstance(err, str)
+
+        self.assertIn('SUPER_UNIQUE_VAR=The answer is 42', out)
+
+    def test_binary(self):
+        env_vars = {'SUPER_UNIQUE_VAR': 'The answer is 42'}
+
+        out, err = processutils.execute('/usr/bin/env',
+                                        env_variables=env_vars,
+                                        binary=True)
+        self.assertIsInstance(out, bytes)
+        self.assertIsInstance(err, bytes)
 
         self.assertIn(b'SUPER_UNIQUE_VAR=The answer is 42', out)
 
@@ -314,6 +391,8 @@ grep foo
                                 'something')
 
         self.assertEqual(38, err.exit_code)
+        self.assertIsInstance(err.stdout, six.text_type)
+        self.assertIsInstance(err.stderr, six.text_type)
         self.assertIn('onstdout --password="***"', err.stdout)
         self.assertIn('onstderr --password="***"', err.stderr)
         self.assertEqual(err.cmd, ' '.join([tmpfilename,
@@ -321,6 +400,72 @@ grep foo
                                             'something']))
         self.assertNotIn('secret', str(err))
 
+    def execute_undecodable_bytes(self, out_bytes, err_bytes,
+                                  exitcode=0, binary=False):
+        if six.PY3:
+            code = ';'.join(('import sys',
+                             'sys.stdout.buffer.write(%a)' % out_bytes,
+                             'sys.stdout.flush()',
+                             'sys.stderr.buffer.write(%a)' % err_bytes,
+                             'sys.stderr.flush()',
+                             'sys.exit(%s)' % exitcode))
+        else:
+            code = ';'.join(('import sys',
+                             'sys.stdout.write(%r)' % out_bytes,
+                             'sys.stdout.flush()',
+                             'sys.stderr.write(%r)' % err_bytes,
+                             'sys.stderr.flush()',
+                             'sys.exit(%s)' % exitcode))
+
+        return processutils.execute(sys.executable, '-c', code, binary=binary)
+
+    def check_undecodable_bytes(self, binary):
+        out_bytes = b'out: ' + UNDECODABLE_BYTES
+        err_bytes = b'err: ' + UNDECODABLE_BYTES
+        out, err = self.execute_undecodable_bytes(out_bytes, err_bytes,
+                                                  binary=binary)
+        if six.PY3 and not binary:
+            self.assertEqual(out, os.fsdecode(out_bytes))
+            self.assertEqual(err, os.fsdecode(err_bytes))
+        else:
+            self.assertEqual(out, out_bytes)
+            self.assertEqual(err, err_bytes)
+
+    def test_undecodable_bytes(self):
+        self.check_undecodable_bytes(False)
+
+    def test_binary_undecodable_bytes(self):
+        self.check_undecodable_bytes(True)
+
+    def check_undecodable_bytes_error(self, binary):
+        out_bytes = b'out: password="secret1" ' + UNDECODABLE_BYTES
+        err_bytes = b'err: password="secret2" ' + UNDECODABLE_BYTES
+        exc = self.assertRaises(processutils.ProcessExecutionError,
+                                self.execute_undecodable_bytes,
+                                out_bytes, err_bytes, exitcode=1,
+                                binary=binary)
+
+        out = exc.stdout
+        err = exc.stderr
+        out_bytes = b'out: password="***" ' + UNDECODABLE_BYTES
+        err_bytes = b'err: password="***" ' + UNDECODABLE_BYTES
+        if six.PY3:
+            # On Python 3, stdout and stderr attributes of
+            # ProcessExecutionError must always be Unicode
+            self.assertEqual(out, os.fsdecode(out_bytes))
+            self.assertEqual(err, os.fsdecode(err_bytes))
+        else:
+            # On Python 2, stdout and stderr attributes of
+            # ProcessExecutionError must always be bytes
+            self.assertEqual(out, out_bytes)
+            self.assertEqual(err, err_bytes)
+
+    def test_undecodable_bytes_error(self):
+        self.check_undecodable_bytes_error(False)
+
+    def test_binary_undecodable_bytes_error(self):
+        self.check_undecodable_bytes_error(True)
+
 
 class ProcessExecutionErrorLoggingTest(test_base.BaseTestCase):
     def setUp(self):
@@ -426,21 +571,23 @@ class FakeSshChannel(object):
         return self.rc
 
 
-class FakeSshStream(six.StringIO):
+class FakeSshStream(six.BytesIO):
     def setup_channel(self, rc):
         self.channel = FakeSshChannel(rc)
 
 
 class FakeSshConnection(object):
-    def __init__(self, rc):
+    def __init__(self, rc, out=b'stdout', err=b'stderr'):
         self.rc = rc
+        self.out = out
+        self.err = err
 
     def exec_command(self, cmd):
-        stdout = FakeSshStream('stdout')
+        stdout = FakeSshStream(self.out)
         stdout.setup_channel(self.rc)
-        return (six.StringIO(),
+        return (six.BytesIO(),
                 stdout,
-                six.StringIO('stderr'))
+                six.BytesIO(self.err))
 
 
 class SshExecuteTestCase(test_base.BaseTestCase):
@@ -455,9 +602,70 @@ class SshExecuteTestCase(test_base.BaseTestCase):
                           None, 'ls', process_input='important')
 
     def test_works(self):
-        o, e = processutils.ssh_execute(FakeSshConnection(0), 'ls')
-        self.assertEqual('stdout', o)
-        self.assertEqual('stderr', e)
+        out, err = processutils.ssh_execute(FakeSshConnection(0), 'ls')
+        self.assertEqual('stdout', out)
+        self.assertEqual('stderr', err)
+        self.assertIsInstance(out, six.text_type)
+        self.assertIsInstance(err, six.text_type)
+
+    def test_binary(self):
+        o, e = processutils.ssh_execute(FakeSshConnection(0), 'ls',
+                                        binary=True)
+        self.assertEqual(b'stdout', o)
+        self.assertEqual(b'stderr', e)
+        self.assertIsInstance(o, bytes)
+        self.assertIsInstance(e, bytes)
+
+    def check_undecodable_bytes(self, binary):
+        out_bytes = b'out: ' + UNDECODABLE_BYTES
+        err_bytes = b'err: ' + UNDECODABLE_BYTES
+        conn = FakeSshConnection(0, out=out_bytes, err=err_bytes)
+
+        out, err = processutils.ssh_execute(conn, 'ls', binary=binary)
+        if six.PY3 and not binary:
+            self.assertEqual(out, os.fsdecode(out_bytes))
+            self.assertEqual(err, os.fsdecode(err_bytes))
+        else:
+            self.assertEqual(out, out_bytes)
+            self.assertEqual(err, err_bytes)
+
+    def test_undecodable_bytes(self):
+        self.check_undecodable_bytes(False)
+
+    def test_binary_undecodable_bytes(self):
+        self.check_undecodable_bytes(True)
+
+    def check_undecodable_bytes_error(self, binary):
+        out_bytes = b'out: password="secret1" ' + UNDECODABLE_BYTES
+        err_bytes = b'err: password="secret2" ' + UNDECODABLE_BYTES
+        conn = FakeSshConnection(1, out=out_bytes, err=err_bytes)
+
+        out_bytes = b'out: password="***" ' + UNDECODABLE_BYTES
+        err_bytes = b'err: password="***" ' + UNDECODABLE_BYTES
+
+        exc = self.assertRaises(processutils.ProcessExecutionError,
+                                processutils.ssh_execute,
+                                conn, 'ls',
+                                binary=binary, check_exit_code=True)
+
+        out = exc.stdout
+        err = exc.stderr
+        if six.PY3:
+            # On Python 3, stdout and stderr attributes of
+            # ProcessExecutionError must always be Unicode
+            self.assertEqual(out, os.fsdecode(out_bytes))
+            self.assertEqual(err, os.fsdecode(err_bytes))
+        else:
+            # On Python 2, stdout and stderr attributes of
+            # ProcessExecutionError must always be bytes
+            self.assertEqual(out, out_bytes)
+            self.assertEqual(err, err_bytes)
+
+    def test_undecodable_bytes_error(self):
+        self.check_undecodable_bytes_error(False)
+
+    def test_binary_undecodable_bytes_error(self):
+        self.check_undecodable_bytes_error(True)
 
     def test_fails(self):
         self.assertRaises(processutils.ProcessExecutionError,
@@ -465,13 +673,13 @@ class SshExecuteTestCase(test_base.BaseTestCase):
 
     def _test_compromising_ssh(self, rc, check):
         fixture = self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
-        fake_stdin = six.StringIO()
+        fake_stdin = six.BytesIO()
 
         fake_stdout = mock.Mock()
         fake_stdout.channel.recv_exit_status.return_value = rc
-        fake_stdout.read.return_value = 'password="secret"'
+        fake_stdout.read.return_value = b'password="secret"'
 
-        fake_stderr = six.StringIO('password="foobar"')
+        fake_stderr = six.BytesIO(b'password="foobar"')
 
         command = 'ls --password="bar"'
 
